#ifndef CONNECTION_MANAGER_H
#define CONNECTION_MANAGER_H

#include "udp.h"
#include "tcp.h"
#include "connection.h"
#include <set>
#include <map>
#include <queue>

namespace Librescue {
  class ConnectionManager; // Forward declaration

  class Listener {
  protected:
	Connection* connection;
	ConnectionManager* manager;
	pthread_t thread;
	pthread_mutex_t lock;

	// Override this if the subclass needs to do something special (e.g. tell the connection manager whether it should send to the from address via TCP or UDP).
	virtual void handleInput(const Address& from, const Input& buffer) = 0;

  public:
	Listener(Connection* c, ConnectionManager* manager);
	virtual ~Listener();

	void start();
	void stop();
	void poll();

  private:
	bool started;

	Listener(const Listener& l);
	Listener& operator=(const Listener& rhs);
  };

  class UDPListener: public Listener {
  protected:
	virtual void handleInput(const Address& from, const Input& buffer);

  public:
	UDPListener(Connection* c, ConnectionManager* manager);
	virtual ~UDPListener();
  };

  class TCPListener: public Listener {
  protected:
	virtual void handleInput(const Address& from, const Input& buffer);

  public:
	TCPListener(Connection* c, ConnectionManager* manager);
	virtual ~TCPListener();
  };

  typedef struct InputInfo {
	Input input;
	Address from;
  } InputInfo;

  typedef struct PollInfo {
	Listener* listener;
	ConnectionManager* manager;
  } PollInfo;

  class ConnectionManager : public TcpServer::TCPConnectionCallback {
	friend class Listener;
	friend class UDPListener;
	friend class TCPListener;

  private:
	int m_udpWait;
	int m_udpSize;
	//	bool m_listenOnUDPSendPort;
	LongUDPConnection* m_udp; // The connection to use when sending LongUDP
	typedef std::list<Listener*> ListenerList;
	typedef std::set<Address> AddressSet;
	typedef std::map<Address,TcpConnection*> AddressToTCPConnection;
	typedef std::list<TcpServer*> TCPServerList;
	ListenerList m_listeners; // Connections that are listening for incoming data
	AddressSet m_udpTargets; // Targets that should be sent LongUDP data.
	AddressSet m_tcpTargets; // Targets that should be sent TCP data.
	AddressToTCPConnection m_tcpOpen; // Open TCP connections
	TCPServerList m_tcpServers; // TCP servers that are listening for connections
	bool started; // Whether the manager has been started or not

	// Send a message to a target
	virtual bool sendUDP(const Output& buffer, const Address& address);
	virtual bool sendTCP(const Output& buffer, const Address& address);

	virtual LongUDPConnection* openLongUDPConnection(const Address& target, int timeout = 60);
	virtual void closeLongUDPConnection(LongUDPConnection*);

	// Open a new TCP connection to a particular address - this will create a new object, so don't forget to call closeTCPConnection later on
	virtual TcpConnection* openTCPConnection(const Address& address);
	// Close an existing TCP connection - this will delete the connection object
	virtual void closeTCPConnection(TcpConnection*);

	Listener* createListener(LongUDPConnection* connection);
	Listener* createListener(TcpConnection* connection);
	void destroyListener(Listener* listener);

	virtual void addData(InputInfo info);

	std::queue<InputInfo> m_data;
	pthread_mutex_t m_dataLock;
	pthread_cond_t m_dataCondition;

	// No copying allowed
	ConnectionManager(const ConnectionManager& rhs);
	ConnectionManager& operator=(const ConnectionManager& rhs);

  public:
	ConnectionManager();
	virtual ~ConnectionManager();

	// Start the connection manager
	virtual void start();
	// Stop the connection manager
	virtual void stop();

	// Send some data to an address. Returns true iff the message was sent successfully.
	virtual bool send(const Output& buffer, const Address& address);
	// Receive data and store in a buffer. If "from" is not null then the address the data came from will be stored in "from". This function will block until data is available, or until "timeout" milliseconds have elapsed. If timeout is zero then this function will not block (this is the default), if timeout is less than zero then it will not timeout. Returns true iff data has been received.
	virtual bool receive(Input& buffer, Address* from, int timeout = 0);
	// Find out if any data is available for reading. If timeout is zero then this function will not block (this is the default), if timeout is less than zero then it will not timeout. Returns true iff data has been received.
	virtual bool isDataAvailable(int timeout = 0);

	// TCP
	// Instruct the connection manager to listen for TCP connections on a particular port
	virtual void listenTCP(int port);
	// In the future send all data to the given target via TCP
	virtual void sendViaTCP(const Address& target);
	// Callback function for TCPServers
	virtual void incomingTCPConnection(TcpConnection*, const Address& from);

	// UDP
	virtual void setUDPSize(int size);
	virtual void setUDPWait(int wait);
	// Instruct the connection manager to listen for UDP connections on a particular address and port.
	// Address: The network byte order address to bind to.
	// Port: The network byte order port to bind to (can be zero).
	// Timeout: If a LongUDP packet is not completely received within this number of seconds then consider the packet to be timed out.
	virtual void listenUDP(int port, const char* address = 0, int timeout = 60);
	// In the future send all data to the given target via UDP
	virtual void sendViaUDP(const Address& target);
	// Should we listen to the port we use for sending data?
	//	virtual void setListenOnUDPSendPort(bool b);
  };  

  // Check if a Connection has data available. If it has, pull the data out and put it in the queue. This expects an argument of type PollInfo*.
  void* poll(void* info);

  inline bool before(struct timeval* first, struct timespec* last);
}

#endif
